/*
    unc_fsx.c
    Copyright 1998, 1999 by Richard Dawe

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    ---

    This is the UNC filename file-system extension code. It uses the
    lm_NetUse*() functions to map shares to drive letters as needed, and then
    modifies filenames to use these mapped shares.
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <sys/fsext.h>
#include <sys/vfs.h>

#include <sys/movedata.h>
#include <sys/segments.h>

#include "lm_use.h"
#include "unc_fsx.h"

/* --- Globals --- */

int unc_fsext_inited = 0;

UNC_FSEXT_INFO *unc_fsext_info_chain = NULL;

/* --- Functions --- */

int isUNC (char *file);

int findFreeFS (void);
int isUNCMapped (char *file);
int isDeviceMapped (char *file);

char *getUNCMapping (char *file);
int getUNCMappings (void);
int mapUNCtoLocal (char *file, char *newdevname);

char *getUNCShareName (char *file);

int addUNCMapping (char *devname, char *remotename);
int removeUNCMapping (char *devname, int safeflag);
int removeUNCMappingSafely (char *devname) { return(removeUNCMapping(devname, 1)); }

void setUNCMappingsStartFlag (void);
int increaseUNCMappingCount (char *devname);
int decreaseUNCMappingCount (char *devname);

int unc_fsext (__FSEXT_Fnumber funcnum, int *rv, va_list args);
void unc_fsext_uninit_wrapper (void);

/* ------------------
   - unc_fsext_init -
   ------------------ */

int unc_fsext_init (void)
{
    if (unc_fsext_inited) return(unc_fsext_inited);

    /* Get current UNC -> device mappings */
    getUNCMappings();

    /* Set the bits that indicate these were here on start-up */
    setUNCMappingsStartFlag();

    /* NB: Return value doesn't indicate success! */
    __FSEXT_add_open_handler(unc_fsext);

    /* Tidy up on exit */
    atexit(unc_fsext_uninit_wrapper);

    /* Done */
    unc_fsext_inited = 1;

    /* OK */
    return(unc_fsext_inited);
}

/* --------------------
   - unc_fsext_uninit -
   -------------------- */

void unc_fsext_uninit_wrapper (void) { unc_fsext_uninit(); }

int unc_fsext_uninit (void)
{
    UNC_FSEXT_INFO *lfi = unc_fsext_info_chain;
    UNC_FSEXT_INFO *prev_lfi = NULL;

    /* Remove all shares that weren't present before */
    while (lfi != NULL) {
        /* 0 = don't force it */
        if ( (!lfi->startflag) && (*lfi->devname != '\0') )
            lm_NetUseDel(lfi->devname, 0);

        prev_lfi = lfi;
        lfi = lfi->next;
        free(prev_lfi);
    }

    /* OK */
    unc_fsext_inited = 0;
    return(1);
}

/* -------------
   - unc_fsext -
   ------------- */

int unc_fsext (__FSEXT_Fnumber funcnum, int *rv, va_list args)
{
    char *file, *devname, *remotename;
    /*int mode, perm;*/
    int freefs, ret;

    /* Assume the call isn't emulated first. */
    ret = 0;

    switch (funcnum) {
        case __FSEXT_open:
            /* Get params */
            file = va_arg(args, char *);

            /* UNC file? */
            if (isUNC(file)) {
                if (isUNCMapped(file)) {
                    /* Mapping already present - just use it */
                    devname = getUNCMapping(file);

                    if (!mapUNCtoLocal(file, devname)) {
                        *rv = -1; ret = 1; break;
                    }

                    increaseUNCMappingCount(devname);
                } else {
                    /* Need to create a mapping */
                    freefs = findFreeFS();
                    if (freefs == -1) { *rv = -1; ret = 1; break; }

                    /* Need space for device name */
                    devname = malloc(3 * sizeof(char));
                    if (devname == NULL) { *rv = -1; ret = 1; break; }
                    sprintf(devname, "%c:", 'A' + freefs - 1);

                    /* Get remote name */
                    remotename = getUNCShareName(file);
                    if (remotename == NULL) { *rv = -1; ret = 1; break; }

                    /* Create the mapping */
                    if (!addUNCMapping(devname, remotename)) {
                        *rv = -1; ret = 1; break;
                    }

                    /* Now use it */
                    if (!mapUNCtoLocal(file, devname)) {
                        *rv = -1; ret = 1; break;
                    }

                    increaseUNCMappingCount(devname);
                    free(devname);
                }
            }

            /* Pretend we haven't done anything so file op's continue as
               normal. Teehee. */
            break;

        /* Maybe unmap files on close? Seems a bit risky really. */
        case __FSEXT_close:
            /*handle = va_arg(args, int *);

            devname = getDeviceName(file);
            decreaseUNCMappingCount(devname);
            removeUNCMappingSafely(devname);*/
            break;

        /* All calls should work OK after initial mapping. The DOS redirector
           should handle it OK. */
        default:
            break;
    }

    /* Done */
    return(ret);
}

/* ---------
   - isUNC -
   --------- */

/* Strictly UNC names should use backslashes, but since we're using DJGPP, I
   guess a bit more flexiblility is a good idea. Besides, people might try
   using forward slashes too... */

int isUNC (char *file)
{
    char *p, *p2;

    /* Must start \\ or // */
    if ( (file[0] != '\\') && (file[0] != '/') ) return(0);
    if ( (file[1] != '\\') && (file[1] != '/') ) return(0);

    /* Must be \\<host>\<share>\... form (or with /s) => two more slashes
       somewhere */
    p = strchr(&file[2], '\\');
    if (p == NULL) p = strchr(&file[2], '/');
    if (p == NULL) return(0);

    p2 = p;
    p = strchr(p + 1, '\\');
    if (p == NULL) p = strchr(p2 + 1, '/');
    if (p == NULL) return(0);

    /* Yep */
    return(1);
}

/* --------------
   - findFreeFS -
   -------------- */

/* Find a free drive letter & return ascii number 1-26 on success, -1 on
   failure (NB: 0 corresponds to current drive!). */

int findFreeFS (void)
{
    int i, ret, freefs;
    char devname[4];
    struct statfs sfs;

    for (freefs = -1, i = 26; i > 0; i--) {
        /* Look at the root dir */
        sprintf(devname, "%c:/", 'A' + i - 1);
        bzero(&sfs, sizeof(sfs));
        ret = statfs(devname, &sfs);

        if (ret != 0) {
            /* TODO: This may be incredibly dodgy! */
            /* Failed => free? */
            freefs = i;
            break;
        }
    }

    /* Done */
    return(freefs);
}

/* ---------------
   - isUNCMapped -
   --------------- */
 
int isUNCMapped (char *file)
{
    char *remotename;
    UNC_FSEXT_INFO *lfi = unc_fsext_info_chain;

    if (lfi == NULL) return(0);             /* No info => not mapped */

    /* Get the UNC path name */
    remotename = getUNCShareName(file);
    if (remotename == NULL) return(0);      /* No remote => not mapped */

    /* Check the rest */
    while(lfi != NULL) {                    
        if (   (stricmp(remotename, lfi->remotename) == 0)
            && (*lfi->devname != '\0') ) {
            free(remotename);
            return(1);
        }

        /* Next! */
        lfi = lfi->next;
    }

    /* Failed */
    free(remotename);
    return(0);
}

/* -----------------
   - getUNCMapping -
   ----------------- */

/* This returns the device name, e.g. Z:, corresponding to a device->share
   mapping. */

char *getUNCMapping (char *file)
{
    char *remotename;
    UNC_FSEXT_INFO *lfi = unc_fsext_info_chain;

    /* No info => not mapped */
    if (lfi == NULL) return(NULL);

    /* Get the remote name */
    remotename = getUNCShareName(file);
    if (remotename == NULL) return(NULL);

    /* Check the rest */
    while(lfi != NULL) {                    
        if (stricmp(remotename, lfi->remotename) == 0) return(lfi->devname);

        /* Next! */
        lfi = lfi->next;
    }

    return(NULL);
}

/* -----------------
   - mapUNCtoLocal -
   ----------------- */

int mapUNCtoLocal (char *file, char *devname)
{
    char *p, *file2;

    /* Check the parameters */
    if ( (file == NULL) || (devname == NULL) ) return(0);
    if (!isUNC(file)) return(0);

    /* Work on a copy */
    file2 = strdup(file);

    /* Convert all '/'s to '\'s */
    for (p = file2; p < file2 + strlen(file2); p++) { if (*p == '/') *p = '\\'; }

    /* Find the end of the \\host\share part */
    p = strchr(file2, '\\');
    p += 2;                     /* This is OK - checked for UNC path already */
    p = strchr(p, '\\');
    p++;
    p = strchr(p, '\\');
    p++;

    /* Place converted version in file */
    sprintf(file, "%s\\%s", devname, p);

    /* Done */
    free(file2);    
    return(1);
}

/* ------------------
   - getUNCMappings -
   ------------------ */

int getUNCMappings (void)
{
    int i, count, ret;
    LM_USE_INFO_0 mapped[26];   /* Buffer for 26 mappings  */
    char unc[262];              /* Max UNC file length buf */
    UNC_FSEXT_INFO *lfi;        /* FSEXT info pointer      */

    /* Enumerate current shares */
    count = ret = lm_NetUseEnum(0, (void *) mapped, sizeof(mapped));

    /* Failed */
    if (ret == -1) return(0);

    /* OK */
    if (ret > 0) {
        for (i = 0; i < count; i++) {
            /* Get the UNC name from DOS */
            movedata(lm_bufsel,  mapped[i].unc_off, _my_ds(), (int) unc, 262);

            /* Does this mapping already have info struct? */
            lfi = unc_fsext_info_chain;

            while (lfi != NULL) {
                if (   (stricmp(lfi->devname, mapped[i].devname) == 0)
                    && (stricmp(lfi->remotename, unc)            == 0) )
                    break;

                lfi = lfi->next;        /* Next! */
            }

            if (lfi != NULL) continue;  /* Yes, already there */

            /* Need to create mapping info */
            lfi = unc_fsext_info_chain;

            if (lfi == NULL) {
                lfi = unc_fsext_info_chain = malloc(sizeof(UNC_FSEXT_INFO));
                if (lfi == NULL) return(0);
            } else {
                lfi->next = malloc(sizeof(UNC_FSEXT_INFO));
                if (lfi->next == NULL) return(0);
                lfi = lfi->next;
            }

            bzero(lfi, sizeof(UNC_FSEXT_INFO));
            lfi->devname    = strdup(mapped[i].devname);
            lfi->remotename = strdup(unc);
            lfi->next       = NULL;
            lfi->startflag  = 0;
        }
    }

    /* OK */
    return(1);
}

/* ---------------------------
   - setUNCMappingsStartFlag -
   --------------------------- */

/* This could be used to determine which shares were used at start-up and
   which were mapped by the FSEXT. */

void setUNCMappingsStartFlag (void)
{
    UNC_FSEXT_INFO *lfi = unc_fsext_info_chain;

    while (lfi != NULL) {
        lfi->startflag = 1;
        lfi = lfi->next;
    }
}

/* -------------------
   - getUNCShareName -
   ------------------- */

char *getUNCShareName (char *file)
{
    char *remotename, *p;

    /* Need a UNC path */
    if (!isUNC(file)) return(NULL);

    /* Work on a copy */
    remotename = strdup(file);
    if (remotename == NULL) return(NULL);

    p = remotename + 2;                             /* Skip \\ or //   */
    while ( (*p != '\\') && (*p != '/') ) { p++; }  /* Skip host       */
    p++;                                            /* Skip \ or /     */
    while ( (*p != '\\') && (*p != '/') ) { p++; }  /* Skip share name */
    *p = '\0';                                      /* Terminate now!  */

    return(remotename);
}

/* -----------------
   - addUNCMapping -
   ----------------- */

int addUNCMapping (char *devname, char *remotename)
{
    int ret;
    UNC_FSEXT_INFO *lfi;

    /* Only work with disks */
    ret = lm_NetUseAdd(devname, remotename, NULL, LM_USE_TYPE_DISK);
    if (ret == -1) return(0);   /* Failed! */

    /* Add an info structure to the chain */
    if (unc_fsext_info_chain == NULL) {
        lfi = unc_fsext_info_chain = malloc(sizeof(UNC_FSEXT_INFO));
        if (lfi == NULL) return(0);
    } else {
        lfi = unc_fsext_info_chain;
        while (lfi->next != NULL) { lfi = lfi->next; }
        lfi->next = malloc(sizeof(UNC_FSEXT_INFO));
        if (lfi->next == NULL) return(0);
        lfi = lfi->next;
    }

    /* Fill the info structure */
    bzero(lfi, sizeof(lfi));
    lfi->devname    = strdup(devname);
    lfi->remotename = strdup(remotename);
    lfi->next       = NULL;
    lfi->startflag  = 0;

    /* OK */
    return(1);
}

/* --------------------
   - removeUNCMapping -
   -------------------- */

int removeUNCMapping (char *devname, int safeflag)
{
    UNC_FSEXT_INFO *lfi, *prev_lfi;

    lfi = unc_fsext_info_chain;
    if (lfi == NULL) return(1);     /* No chain => success */

    /* Find the device in the list */
    while (lfi != NULL) {
        if (stricmp(devname, lfi->devname) == 0) break;
        prev_lfi = lfi;
        lfi = lfi->next;
    }

    if (lfi == NULL) return(1);     /* Not found => success */

    /* Don't remove if still used */
    if (safeflag && (lfi->count > 0)) return(0);

    /* Now remove it */    
    prev_lfi->next = lfi->next;
    free(lfi->remotename);
    free(lfi->devname);
    free(lfi);
    return(1);
}

/* ---------------------------
   - increaseUNCMappingCount -
   --------------------------- */

int increaseUNCMappingCount (char *devname)
{
    UNC_FSEXT_INFO *lfi;

    lfi = unc_fsext_info_chain;
    if (lfi == NULL) return(-1);     /* No chain => fail */

    /* Find the device in the list */
    while (lfi != NULL) {
        if (stricmp(devname, lfi->devname) == 0) break;
        lfi = lfi->next;
    }

    if (lfi == NULL) return(-1);     /* No mapping => fail */

    (lfi->count)++;

    return(lfi->count);
}

/* ---------------------------
   - decreaseUNCMappingCount -
   --------------------------- */

int decreaseUNCMappingCount (char *devname)
{
    UNC_FSEXT_INFO *lfi;

    lfi = unc_fsext_info_chain;
    if (lfi == NULL) return(-1);     /* No chain => fail */

    /* Find the device in the list */
    while (lfi != NULL) {
        if (stricmp(devname, lfi->devname) == 0) break;
        lfi = lfi->next;
    }

    if (lfi == NULL) return(-1);     /* No mapping => fail */

    if (lfi->count > 0) (lfi->count)--;
    return(lfi->count);
}

/* ------------------
   - isDeviceMapped -
   ------------------ */

int isDeviceMapped (char *file)
{
    char devname[3];
    UNC_FSEXT_INFO *lfi;

    /* Not of form '<drive letter>:' */
    if (file[1] != ':') return(0);

    /* Copy the device name */
    devname[0] = file[0];
    devname[1] = devname[1];
    devname[2] = '\0';    

    /* Walk the info chain */
    lfi = unc_fsext_info_chain;
    if (lfi == NULL) return(0);      /* No chain => fail */

    /* Find the device in the list */
    while (lfi != NULL) {
        if (stricmp(devname, lfi->devname) == 0) break;
        lfi = lfi->next;
    }

    return( (lfi == NULL) ? 0 : 1);
}
